%% /u/beebe/src/lex/cweb/expr.w, Tue May 28 20:18:53 1991
%% Edit by Nelson H. F. Beebe <beebe@@rios.math.utah.edu>

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% I don't like the default long lines from
%% cwebmac.tex; include dimension changes from
%% LaTeX's art10.doc that make the output more
%% readable.

\pagewidth=345pt

% Adjust \pageheight to 43 lines/page with
% 10-on-12 type
\pageheight=12pt
\multiply \pageheight by 43
\fullpageheight=\pageheight
\advance \fullpageheight by 24pt

% Adjust \hoffset to (paperwidth - \pagewidth)/2
\hoffset=8.5in
\advance\hoffset by -2in
\advance\hoffset by -\pagewidth
\divide\hoffset by 2
\setpage            % update everything that
                    % depends on the above

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Define some miscellaneous macros to help the
%% formatting Much of this is standard boilerplate
%% that can be used in other WEB programs, and
%% therefore, can be relegated to a private macro
%% file.

\def\ifundefined#1{
    \expandafter\ifx\csname#1\endcsname\relax
    }

\ifundefined{METAFONT}
        \font\mf=logo10
        \hyphenchar\mf=-1
        \def\METAFONT{{\mf METAFONT}}
\fi

\let\MF=\METAFONT

\let\em=\it

\def\sloppy{
%   \tolerance 10000
    \tolerance 8000
    \hfuzz .5pt
    \vfuzz .5pt}

\def\sloppypar{\par\sloppy}

\ifundefined{emergencystretch}
    \relax
\else
    \emergencystretch=3pt
\fi

\def\ctangle{\.{ctangle}}
\def\cweave{\.{cweave}}
\def\cweb{\.{cweb}}
\def\WEB{\.{WEB}}

\def\langle{{\tt\char`\<}}

\def\rangle{{\tt\char`\>}}

\def\vbar{{\tt\char`\|}}

% Redefine C operators to use conventional
% symbols, rather than mathematical ones

\def\I{\mathbin{!=}}
\def\V{\mathbin{\mid}}
\def\W{\mathbin{\char`\&\char`\&}}
\def\XOR{`\^}
\def\NULL{\.{NULL}}

\input tablisting.sty

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\def\title{EXPR}  % used in the page headers

\def\topofcontents{\null\vfill
  \centerline{\titlefont
    {\ttitlefont EXPR}---A
    Simple Expression Parser}
  \vskip 15pt
  \centerline{Nelson H.F. Beebe}
  \centerline{Center for Scientific Computing}
  \centerline{Department of Mathematics}
  \centerline{South Physics Building}
  \centerline{University of Utah}
  \centerline{Salt Lake City, UT 84112}
  \centerline{USA}
  \centerline{Tel: (801) 581-5254}
  \centerline{FAX: (801) 581-4148}
  \centerline{Email: \.{<beebe@@math.utah.edu>}}
  \vskip 15pt
  \centerline{11 April 1991}
  \vfill
  }

@* Introduction.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

This program illustrates Donald Knuth's {\em
literate programming\/} model using Silvio Levy's
\cweb{} system.  The document you are now reading
is produced from the \cweb{} file \.{expr.w} by
running \cweave{} to make a \TeX{} file,
\.{expr.tex}, from it.  The \Cee{} program itself
is generated by running \ctangle{} on \.{expr.w}
to make \.{expr.c} which can be compiled, linked,
and run in the normal fashion.\sloppypar

Literate programming is based on the idea that
explanatory paragraphs followed by short code
sections can make a program much easier to
understand.  The \WEB{} system, of which \cweb{}
is just one implementation, provides for the
generation of the document and program from a
single master file, possibly modified by a {\em
change file}.  Change files permit local
customizations and software ports without ever
needing to modify the base file.

\WEB{} provides a number of pleasant facilities,
including nice typography for the program text,
and extensive cross-referencing of code sections
and variables.

Of course, the program author has to work harder,
because good typography encourages efforts toward
good exposition, and because two
jobs---documentation and programming---are being
carried out at the same time.  The opportunity for
syntax errors is tripled---such errors can be
found by \WEB{}, \TeX{}, and the language
compiler.\sloppypar

On the other hand, the experience of using \WEB{}
for the \TeX{} project, where it has been used to
program \TeX{}, \MF{}, and numerous utilities,
totaling over 50~000 lines of compiled code, has
been that {\em programming\/} errors are
significantly reduced.  The discipline of
simultaneous description for a human reader, and
programming for a compiler, tends to clarify the
author's thoughts and helps to get it right at an
earlier stage.

To get the whole story about \WEB{}, consult
E.~Wayne Sewell's book {\em Weaving a Program:
Literate Programming in {\WEB}}, Van Nostrand
Reinhold, 1989.

@* Using the make utility.

The \UNIX{} \.{make} program facilitates using
\WEB{}, since the multiple steps needed to run
\ctangle{}, \cweave{}, \TeX{}, and the C compiler
can be written down as rules in the \.{Makefile}.

Here is the \.{Makefile} used for this program:

\tablisting{Makefile}

@* The expression grammar.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
The program is a simple double-precision
calculator.  It evaluates expressions based on
this grammar:

\settabs \+E' & : & number\cr   % sample line
\+        E & : &E + T\cr
\+           & \vbar & E - T\cr
\+           & \vbar & T\cr

\+        T & : &T * F\cr
\+           & \vbar & T / F\cr
\+           & \vbar & F\cr

\+        F & : &number\cr
\+           & \vbar & ( E )\cr

\noindent
Eliminating left recursion leads to this version
which we implement here in a recursive descent
parser:

\+        E & : &T E'\cr

\+        E' & : &+ T\cr
\+            & \vbar & - T\cr
\+            & \vbar & \langle empty\rangle\cr

\+        T & : &F T'\cr

\+        T' & : &* F\cr
\+            & \vbar & / F\cr
\+            & \vbar & \langle empty\rangle\cr

\+        F & : &number\cr
\+           & \vbar & ( E )\cr

@* Program layout.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Since this is a small program, its structure is
fairly simple:

@c 
@<Top-level macros@>@/
@<Include files@>@/
@<Function prototypes@>@/
@<Global constants@>@/
@<Global variables@>@/
@<Main code@>@/

@ We want this code to work in an old \Cee{}
environment as well as a Standard \Cee{} one, so
as to get the benefit of function prototype
argument checking.  We shall use old-style
function definitions but new-style function
declarations.  The latter become old-style when
the arguments inside the parentheses are dropped.
This is most easily handled by a macro, |ARGS|,
whose expansion depends on whether we are in a
Standard \Cee{} environment or not.

@<Top-level macros@>=
#if    __STDC__
#define ARGS(plist)             plist
#else
#define ARGS(parenthesized_list) ()
#endif

@ Using the |ARGS| macro, we can conveniently give
just one set of function prototypes; the argument
declarations will disappear in those environments
where they are not understood yet.

@<Function prototypes@>=
void    error @, ARGS((void));@/
void    expect @, ARGS((int c));@/
double  expr @, ARGS((void));@/
double  factor @, ARGS((void));@/
void    initialize @, ARGS((void));@/
int     nextnb @, ARGS((void));@/
int     peeknb @, ARGS((void));@/
double  term @, ARGS((void));@/
int     main @, ARGS((int argc, char* argv[]));

@ Virtually all \Cee{} programs need one or more
standard header files, and this one is no
exception.  We'll be doing I/O, character
processing, and a bit of mathematical function
calling, so this set suffices:

@<Include files@>=
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>

@ We also need a type declaration for |strtod()|,
we is available in Standard \Cee{}, but not in
many other implementations.

@<Include files@>=
#if    __STDC__
#include <stdlib.h>
#else
double	strtod @, ARGS((const char *nptr, char **endptr));
#endif

@ Standard \Cee{} defines two standard arguments
for the |exit()| function.  We only need one, and
if it has not already been defined by one of the
included files, we define it here with a value
that works on many, but not all, \Cee{}
implementations.

@<Top-level macros@>=
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 0
#endif

@ We need very little global data, essentially
just an array to store input lines in.  Since we
don't want to unnecessarily restrict the
complexity of expressions input by the user, we
make the array size, |MAXBUF|, larger than almost
anyone would ever be likely to need.  On a small
machine, you could probably reduce it drastically
without it being noticed.

@<Global constants@>=
#define MAXBUF 10240

@ In order to keep the function call interface
clean, we store the input line in a global array,
|buf[]|.  Its use is restricted to the |main()|
program, and the |initialize()| function.  For
efficiency, we access it via a pointer, |pc|,
whose use is restricted to |error()|, |factor()|,
|initialize()|, |nextnb()|, and |peeknb()|.

@<Global variables@>=
char buf[MAXBUF]; char *pc; /* pointer to next
input character */

@ Some interactive programs can usefully have an
introductory banner, such as this one:

@d banner "This is expr [11-Apr-1991], in CWEB\n"

@ \Cee{} |main()| programs tend to look much
alike.  We don't actually need any command-line
arguments here, but we declare them anyway, to
facilitate future enhancements.

The funny business with both |exit()| and |return|
deals with the VAX VMS environment, where the two
have different conventions for the meaning of the
arguments.  In most other \Cee{} implementations,
the |exit()| call could be dropped.

@<Main code@>=
int
main(argc,argv)
int argc;
char* argv[];
{
    (void)printf(banner);
    @<Parse expressions until end-of-file@>
    exit (EXIT_SUCCESS);
    return (EXIT_SUCCESS);
}

@ @<Parse expressions until end-of-file@>=
    (void)printf("expr>");

    while (fgets(buf,MAXBUF,stdin) != (char*)NULL)
    {   /* parse expressions until end-of-file */
        initialize();
        (void)printf("%lg\n",expr());
        if (nextnb())
            error(); /* catch trailing garbage */
        (void)printf("expr>");
        (void)fflush(stdout);
    }

@ Most programs need initialization, and this one
is no exception.  The job is simple: just set the
buffer pointer to point at the start of |buf[]|.
This function is called after the input of each
expression string.

@c @<Initialize@>

@ @<Initialize@>=
void
initialize()
{
    pc = buf;
}


@* The expression parser.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Here is the core of the parser, which implements
the grammar rules

\+        E & : &T E'\cr

\+        E' & : &+ T\cr
\+            & \vbar & - T\cr
\+            & \vbar & \langle empty\rangle\cr

\noindent
Notice that the first rule, E, is handled by the
initialization of the variable |e|.  Evaluation of
the second rule, E', is handled by the |for (;;)|
loop, which returns to the caller of |expr()| when
the next non-blank character is not a sign.

@c @<Parse one expression@>

@ @<Parse one expression@>=
double
expr()
{
    double e = term();

    for (;;)
    {
        if (peeknb() == '+')    /* expr + term */
            nextnb(), e += term();
        else if (peeknb() == '-')/* expr - term */
            nextnb(), e -= term();
        else
            return (e);
    }
}

@ The factor rule, F, in the grammar

\+        F & : &number\cr
\+           & \vbar & ( E )\cr

\noindent
translates to a simple |if|-|else| statement,
depending upon whether lookahead finds an open
parenthesis or not.  The complex job of parsing
of a number is relegated to the standard library
routine, |strtod()|, which fortunately is able to
return in its second argument a pointer to the
next unparsed character.  We use that position to
check for errors (nothing was parsed), and to
advance the input buffer pointer, |pc|.

@c @<Parse a factor@>

@ @<Parse a factor@>=
double
factor()
{
    double e;
    char *p;

    if (peeknb() == '(')        /* ( expr ) */
    {
        nextnb();      /* gobble the left paren */
        e = expr();
        expect(')');
    }
    else                        /* number */
    {
        e = strtod(pc, &p);
        if (p == pc)
            error();
        else
            pc = p;
    }
    return (e);
}

@ The third part of the expression grammar defines
a term, T:

\+        T & : &F T'\cr

\+        T' & : &* F\cr
\+            & \vbar & / F\cr
\+            & \vbar & \langle empty\rangle\cr

\noindent
Like |expr()|, this requires a |for(;;)| loop to
handle the operator lookahead, and return occurs
when no multiplicative operator is found.

@c @<Parse a term@>

@ @<Parse a term@>=
double
term()
{
    double t = factor();

    for (;;)
    {
        if (peeknb() == '*')
            nextnb(), t *= factor();
        else if (peeknb() == '/')
            nextnb(), t /= factor();
        else
            return (t);
    }
}

@ Auxiliary functions.
At points in the grammar where a particular
single-character token parsing is required, we can
call |expect()|.  It gobbles the next character,
and raises an error if that character differs from
the expected one.

@c
void
expect(c)
int c;
{
    if (nextnb() != c)
        error();
}

@ For a simple parser such as this one, we don't
go to the trouble of doing a lexical analysis in
advance.  Instead, we lex `on the fly', and to do
so, we need a function to skip ahead to the next
non-blank character, taking care not to advance
past the end of the input buffer.

Incidentally, if we wanted to add support for
in-line comments, such as from a sharp sign, \#,
until end-of-line, this is the place to do it.  On
encountering a comment start symbol, we would just
flush to the end of the comment.

@c @<Get next non-blank character@>

@ @<Get next non-blank character@>=
int
nextnb()
{
    while (*pc && isspace(*pc))
        ++pc;
    return (*pc ? *pc++ : *pc);
                /* don't advance beyond EOS */
}

@ Lookahead can be handled in one of two ways.  We
can either fetch the next character, and then on
discovering that we don't need it, push it back
into the input stream, and else we can use
|peeknb()| to peek ahead to the next non-blank
character, without actually fetching it.  The
former scheme is more powerful, and can handle
multiple-character lookahead, but the second
approach is simpler, and is adequate for our
needs.

@c @<Peek ahead to next non-blank character@>

@ @<Peek ahead to next non-blank character@>=
int
peeknb()
{
    int c = nextnb();

    pc--;
    return (c);
}

@* Error handling and recovery.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

When an error occurs, we print a message showing
what we have left in the input buffer; the cause
of the error should be at, or just before, that
point.  Error recovery is important in interactive
programs.  Since we are going to return without
telling the parser that an error happened, we
prevent further processing by inserting an
end-of-string symbol at the current buffer
position.

@c @<Error message and recovery@>

@ @<Error message and recovery@>=
void
error()
{
    (void)fprintf(stderr,
        "? Error: remaining text = [%s]\n",pc);
    *pc = '\0';
}

@* Index.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Every identifier used in this program is shown
here, together with a list of the section numbers
where the identifier is used.  Underlined entries
mark sections where the identifier is defined.
One-letter identifiers are indexed only at their
point of definition, since they are likely to be
heavily used for things like array and loop
indexes. 

Although this index does not do so, it is possible
to add additional topics to the index, and in a
longer program, it would be desirable to do so.

\input expr.toc % Make the title page and table of contents

%%%%%%%%%%%%%%%%%%% The End %%%%%%%%%%%%%%%%%%%%
